#!/usr/bin/env python

import argparse
from enum import Enum


def get_parser():
    parser = argparse.ArgumentParser(
        description="Create statistics out of process records file",
    )
    parser.add_argument("file", help="process records file")
    parser.add_argument("-i", "--increment",
                        default=1,
                        type=float,
                        help="size of a single histogram interval")
    parser.add_argument("-r", "--rate",
                        default="linear",
                        choices=["linear", "exponential"],
                        help="set growth rate of histogram intervals")
    parser.add_argument("-f", "--filter", help="comma separated list of histograms to show (age,stime,...)",
                        default="all")
    return parser


def average(data):
    return sum(data) / len(data)


def median(data):
    data = sorted(data)
    length = len(data)
    if length % 2 == 0:
        return (data[length // 2] + data[length // 2 - 1]) / 2
    return data[length // 2]


class HistogramRate(Enum):
    EXPONENTIAL = 1
    LINEAR = 2


def histogram(data, /, start=0.001, step=2, rate=HistogramRate.EXPONENTIAL):
    data = sorted(data)
    left, right = sorted([0, start])
    histogram_data = {}

    i = 0
    while i < len(data):
        if left <= data[i] < right:
            histogram_data[(left, right)] = histogram_data.get(
                (left, right), 0) + 1
            i += 1
        else:
            if rate == HistogramRate.EXPONENTIAL:
                size = right - left
                left = right
                right += step * size
            elif rate == HistogramRate.LINEAR:
                left = right
                right += step
            else:
                raise ValueError(f"Invalid rate type: {rate}")

    return histogram_data


def visualize_histogram(histogram):
    m = max(histogram.values())
    max_width = 60
    pad = max(map(lambda s: len(str(s)), histogram.keys()))

    for (left, right), v in histogram.items():
        bar_width = int(max_width * v / m)
        print(f"{f'[{left}, {right})':{pad}} {bar_width * '█'} {v}")


def retype(data, header):
    if header == "PCOMM":
        return data

    if header in {"age", "utime", "stime", "cutime", "cstime"}:
        return float(data)

    return int(data)


def averageable(header):
    return header in {
        "age",
        "utime",
        "stime",
        "nvcsw",
        "nivcsw",
        "cutime",
        "cstime",
        "inblk",
        "oublk",
        "cinblk",
        "coublk",
    }


def medianable(header):
    return averageable(header)


def histogramable(header):
    return averageable(header) or header in {"UID", "exit", "exitsig"}


def main():
    parser = get_parser()
    args = parser.parse_args()

    if args.rate == "exponential" and args.increment <= 1:
        print("Exponential rate can't have increment <= 1\n")
        parser.print_help()
        return

    log_filename = args.file

    lines = []
    with open(log_filename, 'r') as f:
        lines = f.readlines()

    data = []

    headers = lines[2].strip().split()
    delim = None
    if len(headers) == 1:
        delim = ','
        headers = lines[2].strip(delim).split()

    for line in lines[3:-1]:
        line_tokens = line.strip(delim).rsplit(maxsplit=len(headers) - 1)

        process_data = {}
        for i, v in enumerate(headers):
            process_data[v] = retype(line_tokens[i], v)

        data.append(process_data)

    print(f"Number of processes handled: {len(data)}")
    print(lines[0].strip())
    print(lines[-1].strip())
    print(lines[1].strip())

    operations = [
        (averageable, average, "Average"),
        (medianable, median, "Median"),
        (averageable, min, "Minimal"),
        (averageable, max, "Maximal"),
    ]

    for condition, operation, label in operations:
        print()
        for header in headers:
            if condition(header):
                print(f"{label} {header}: {
                      operation([d[header] for d in data]): .3f}s")

    # histograms
    filtered = list(args.filter.split(','))
    for header in headers:
        if histogramable(header) and (header in filtered or args.filter == "all"):
            print()
            print(f"{header.capitalize()} histogram:")
            h = histogram([d[header] for d in data], start=-1 if not averageable(header) else args.increment,
                          step=args.increment,
                          rate=HistogramRate[args.rate.upper()])
            visualize_histogram(h)


if __name__ == "__main__":
    main()
